/*
 debugutil.cpp

 Copyright (c) 2011 AIST  All Rights Reserved.
 Eclipse Public License v1.0 (http://www.eclipse.org/legal/epl-v10.html)
*/
/*!
 * @file debugutil.cpp
 * @brief デバッグ用関数
 * @date \$Date:: 2011-06-23 14:52:42 +0900 #$
 */
#include <stdio.h>
#include <highgui.h>
#include "debugutil.h"
#include "vectorutil.h"
#include "visionErrorCode.h"
#include "modelpoints.h"

static inline int
roundoff( double d )
{
  return (int)floor( d + 0.5 ); // 四捨五入
}

static void
drawBaseEdge(IplImage* image, const uchar* edge, int grayval)
{
  CvMat EdgeMask;
  cvInitMatHeader(&EdgeMask, image->height, image->width, CV_8UC1, (void *)edge);
  CvScalar color = CV_RGB(grayval, grayval, grayval);
  cvSet(image, color, &EdgeMask);
}

static IplImage*
createDebugImage(const Parameters& parameters, bool color=false, const uchar* edge=NULL, int grayval=255)
{
  CvSize imsize = cvSize(parameters.colsize, parameters.rowsize);
  IplImage* image;

  if (color)
    {
      image = cvCreateImage(imsize, IPL_DEPTH_8U, 3);
    }
  else
    {
      image = cvCreateImage(imsize, IPL_DEPTH_8U, 1);
    }

  if (image)
    {
      cvZero(image);
      // 背景エッジが指定されていたら描く
      if (edge)
        {
          drawBaseEdge(image, edge, grayval);
        }
    }

  return image;
}

static void 
outDebugImage(IplImage* image, const char* name, int id, int display)
{
  char title[PATH_MAX+1] = {0};
  const int extlen = 5; // 拡張子分の余裕
  snprintf(title, sizeof(title)-extlen, "%s%d", name, id);
  if (display)
    {
      cvNamedWindow(title, 1);
      cvShowImage(title, image);
      cvWaitKey(-1);
      cvDestroyWindow(title);
    }
  if (image->nChannels == 1)
    {
      strcat(title, ".pgm");
    }
  else
    {
      strcat(title, ".ppm");
    }
  cvSaveImage(title, image);
  cvReleaseImage(&image);
}

int
drawInputImage(const uchar* src, const Parameters& parameters)
{
  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  IplImage* cvGrayImage = createDebugImage(parameters);
  if (cvGrayImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }
  memcpy(cvGrayImage->imageData, src, parameters.imgsize);
  // 入力画像表示・保存
  outDebugImage(cvGrayImage, "src", id, disp);
  return 0;
}

int
drawEdgeImage(const uchar* edge, const Parameters& parameters)
{
  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  IplImage* cvGrayImage = createDebugImage(parameters);
  if (cvGrayImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }
  memcpy(cvGrayImage->imageData, edge, parameters.imgsize);
  cvConvertScaleAbs(cvGrayImage, cvGrayImage, 255, 0); // 0,1を0,255に変換
  // エッジ２値(0,1)画像表示・保存
  outDebugImage(cvGrayImage, "edge", id, disp);
  return 0;
}

int
drawDetectedLines(const uchar* edge, const Features2D* lineFeatures, const Parameters& parameters)
{  
  // 直線検出結果カラー表示・保存
  // 背景表示画像配列変数名：edge
  // 直線検出結果変数名：lineFeatures

  // 画像メモリを確保して背景エッジを白で描く
  bool color = true;
  IplImage* cvColorImage = createDebugImage(parameters, color, edge);

  if (cvColorImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  // 検出結果の描画
  int* pt = NULL;
  CvPoint  dpt;
  CvScalar red   = CV_RGB(255, 0, 0);
  CvScalar green = CV_RGB(0, 255, 0);
  CvScalar blue  = CV_RGB(0, 0, 255);
  Feature2D* tmpFeature = lineFeatures->feature;
  int i, ii, f, cx, cy, istart, iend;

  for (f = 0; f < lineFeatures->nFeature; f++)
    {
      if (tmpFeature[f].type != ConicType_Line)
        {
          continue;
        }
      if (tmpFeature[f].nTrack >= 0)
        {
          pt = lineFeatures->track[tmpFeature[f].nTrack].Point;
          istart = tmpFeature[f].start;
          iend = tmpFeature[f].end;
          if (istart > iend)
            {
              iend += tmpFeature[f].all;
            }
          for (i = istart; i < iend; i++)
            {
              ii = (i % tmpFeature[f].all) * 2;
              cx = pt[ii];
              cy = pt[ii + 1];
              dpt = cvPoint(cx, cy);
              cvLine(cvColorImage, dpt, dpt, blue);
            }
        }
      // 始点表示
      istart = tmpFeature[f].start * 2;
      iend = tmpFeature[f].end * 2;
      cx = pt[istart];
      cy = pt[istart + 1];
      dpt = cvPoint(cx, cy);
      cvLine(cvColorImage, dpt, dpt, green);
      // 終点表示
      cx = pt[iend];
      cy = pt[iend + 1];
      dpt = cvPoint(cx, cy);
      cvLine(cvColorImage, dpt, dpt, red);
    }

  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  outDebugImage(cvColorImage, "line", id, disp);
  return 0;
}

int
drawDetectedVertices(const Features2D* features, const Parameters& parameters)
{
  // 頂点特徴抽出結果カラー表示・保存
  // 背景表示画像なし
  // 頂点特徴抽出結果変数名：features

  bool color = true;
  IplImage* cvColorImage = createDebugImage(parameters, color);

  if (cvColorImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  Feature2D* tmpFeature = features->feature;
  CvPoint  pt1, pt2, pt3;
  CvScalar green = CV_RGB(0, 255, 0);
  CvScalar white = CV_RGB(255, 255, 255);

  for (int f = 0; f < features->nFeature; f++)
    {
      if (tmpFeature[f].type != ConicType_Hyperbola)
        {
          continue;
        }
      pt1.x = roundoff(tmpFeature[f].center[0]);
      pt1.y = roundoff(tmpFeature[f].center[1]);
      pt2.x = roundoff(tmpFeature[f].startSPoint[0]);
      pt2.y = roundoff(tmpFeature[f].startSPoint[1]);
      pt3.x = roundoff(tmpFeature[f].endSPoint[0]);
      pt3.y = roundoff(tmpFeature[f].endSPoint[1]);
      cvLine(cvColorImage, pt1, pt2, green);
      cvLine(cvColorImage, pt1, pt3, green);
      // 特徴点の描画
      cvLine(cvColorImage, pt1, pt1, white);
    }

  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  outDebugImage(cvColorImage, "Line2Vertex", id, disp);
  return 0;
}

int
drawTrackPoints(const Features2D* features, const Parameters& parameters)
{  
  int i, f, n, cx, cy;
  int* pt = NULL;
  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  int colsize = parameters.colsize;
  int rowsize = parameters.rowsize;
  IplImage* cvTrackImage = createDebugImage(parameters);
  if (cvTrackImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }
  for (f = 0; f < features->nTrack; f++)
    {
      pt = features->track[f].Point;
      for (i = 0; i < features->track[f].nPoint; i++)
        {
          cx = pt[i * 2];
          cy = pt[i * 2 + 1];
          if ((cx >= 0 && cx < colsize) && (cy >= 0 && cy < rowsize))
            {
              n = (cy * colsize + cx);
              cvTrackImage->imageData[n] += 50;
            }
        }
    }
  outDebugImage(cvTrackImage, "track", id, disp);
  return 0;
}

// 楕円の係数と中心から形状を調べる
static void
getEllipseProperty(const double coef[6],   // 楕円係数（入力）
                   const double center[2], // 楕円中心（入力）
                   double rot[2][2],       // 楕円回転行列（出力）
                   double axis[2])         // 楕円の軸長（出力）
{
  double F;

  // 中心を移動した場合の F は
  F = (coef[0] * pow(center[0], 2) + coef[1] * center[0] * center[1]
       + coef[2] * pow(center[1], 2) + coef[3] * center[0] + coef[4] * center[1] + coef[5]);
  // これにより a(x+xc)^2 + b(x+xc)(y+yc) + c(y+yc)^2 + F = 0 の形になる

  // 各軸をx,y軸に一致させた場合のA, Cを求める
  double a[2][2] = { {coef[0], coef[1] / 2.0}, {coef[1] / 2.0, coef[2]} };
  double e[2], ev[2][2];
  double A, C;

  CvMat matA        = cvMat( 2, 2, CV_64FC1, a );
  CvMat eigenvalue  = cvMat( 2, 1, CV_64FC1, e );
  CvMat eigenvector = cvMat( 2, 2, CV_64FC1, ev );
  cvEigenVV( &matA, &eigenvector, &eigenvalue );

  if (e[0] * e[1] < VISION_EPS) // 楕円でない場合
    {
      return;
    }

  // 固有ベクトルを転置して回転行列とする
  CvMat matR = cvMat( 2, 2, CV_64FC1, rot );
  cvTranspose( &eigenvector, &matR );

  A = e[0];
  C = e[1];

  // F による規格化
  if (F != 0.0)
    {
      A /= -(F);
      C /= -(F);
      // これで A x^2 + C y^2 = 1 の標準形
      if ((A <= 0.0) && (C <= 0.0))
        {                       // 虚円もしくは放物線
          axis[0] = 0.0;
          axis[1] = 0.0;
        }
      else
        {
          axis[0] = sqrt(1.0 / A);
          axis[1] = sqrt(1.0 / C);
        }
    }
  else
    {
      // これは A x^2 + C y^2 = 0 の形式
      axis[0] = 0.0;
      axis[1] = 0.0;
    }
  return;
}

int
drawDetectedEllipses(const uchar* edge, const Features2D* features, const Parameters& parameters)
{
  // 楕円検出結果カラー表示・保存
  // 背景表示画像配列変数名：edge

  // 画像メモリを確保して背景エッジを白で描く
  bool color = true;
  IplImage* cvColorImage = createDebugImage(parameters, color, edge);

  if (cvColorImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  // 検出結果の描画
  Feature2D* tmpFeature;
  int f, cx, cy;

  for (f = 0; f < features->nFeature; f ++)
    {
      tmpFeature = &features->feature[f];
      if (tmpFeature->type != ConicType_Ellipse)
        {
          continue;
        }
      cx = roundoff(tmpFeature->center[0]);
      cy = roundoff(tmpFeature->center[1]);
      // 楕円中心の描画
      CvPoint  pt  = cvPoint(cx, cy);
      CvScalar red = CV_RGB(255, 0, 0);
      cvLine(cvColorImage, pt, pt, red);
      double *coef = tmpFeature->coef;
      double *cent = tmpFeature->center;
      double rot[2][2] = {{0}};
      double axis[2] = {0};
      // 楕円形状を得る
      getEllipseProperty( coef, cent, rot, axis );
      pt = cvPoint(cent[0], cent[1]);
      // 左上原点なので角度の符号を反転させる（上下が逆）
      double angle = -atan2(rot[1][0], rot[0][0])*180/M_PI;
      CvSize axes  = cvSize(axis[0], axis[1]);
      CvScalar blue = CV_RGB(0, 0, 255);
      // 楕円の描画
      cvEllipse(cvColorImage, pt, axes, angle, 0, 360, blue);
    }

  int id = parameters.feature2D.id;
  int disp = parameters.dbgdisp;
  outDebugImage(cvColorImage, "ellipse", id, disp);
  return 0;
}

int
drawStereoCorrespondence(const StereoData& stereo, int pairing, const Parameters& parameters)
{
  // 頂点、円中心のステレオ対応結果表示（左画像）・保存
  // 背景表示画像：なし
  // ステレオ対応結果変数名：stereo

  IplImage* cvGrayImage = createDebugImage(parameters);

  if (cvGrayImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  CvPoint  pt;
  CvScalar white = CV_RGB(255, 255, 255);
  Feature2D* Lfeature;
  int ix,iy;

  // 特徴点の描画
  for (int i = 0; i < stereo.numOfconics; i++)
    {
      if (stereo.conics[i].valid != 1)
        {
          continue;
        }
      Lfeature = stereo.conics[i].featureL;
      ix = roundoff(Lfeature->center[0]);
      iy = roundoff(Lfeature->center[1]);
      pt = cvPoint(ix, iy);
      cvLine(cvGrayImage, pt, pt, white);
    }

  outDebugImage(cvGrayImage, "stereo", pairing, parameters.dbgdisp);
  return 0;
}

int
drawStereoVertices(const uchar* edge,
                   const StereoData& stereo,
                   int pairing,
                   const Parameters& parameters,
                   const CameraParam* cameraParam)
{
  // 頂点の３次元復元結果画像（左画像）の表示・保存
  // ３次元復元結果変数：stereo

  // 画像メモリを確保して背景エッジをグレーで描く
  bool color = true;
  IplImage* cvColorImage = createDebugImage(parameters, color, edge, 80);

  if (cvColorImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  int ix, iy;
  CvPoint pt;
  CvScalar red    = CV_RGB(255, 0, 0);
  CvScalar green  = CV_RGB(0, 255, 0);
  CvScalar yellow = CV_RGB(255, 255, 0);
  StereoConic* conic;
  VertexCandidate* vertex;
  Data_2D iPos;
  int i;

  // 特徴点の描画
  for (i = 0; i < stereo.numOfconics; i++)
    {
      conic = &stereo.conics[i];
      vertex = &conic->work.vertex;
      if (conic->featureL->type != ConicType_Hyperbola || vertex->valid != 1)
        {
          continue;
        }
      // 頂点を黄色で描画
      projectXYZ2LR(&iPos, vertex->position, (CameraParam*)cameraParam);
      ix = roundoff(iPos.col);
      iy = roundoff(iPos.row);
      pt = cvPoint(ix, iy);
      cvLine(cvColorImage, pt, pt, yellow);
      // 端点１を緑色で描画
      projectXYZ2LR(&iPos, vertex->endpoint1, (CameraParam*)cameraParam);
      ix = roundoff(iPos.col);
      iy = roundoff(iPos.row);
      pt = cvPoint(ix, iy);
      cvLine(cvColorImage, pt, pt, green);
      // 端点２を赤色で描画
      projectXYZ2LR(&iPos, vertex->endpoint2, (CameraParam*)cameraParam);
      ix = roundoff(iPos.col);
      iy = roundoff(iPos.row);
      pt = cvPoint(ix, iy);
      cvLine(cvColorImage, pt, pt, red);
    }

  outDebugImage(cvColorImage, "vertices3D", pairing, parameters.dbgdisp);

  return 0;
}

int
drawStereoCircles(const uchar* edge,
                  const StereoData& stereo,
                  int pairing,
                  const Parameters& parameters,
                  const CameraParam* cameraParam)
{
  // ３次元円復元結果画像（左画像）の表示・保存
  // ３次元復元結果変数：stereo

  // 画像メモリを確保して背景エッジを白で描く
  bool color = true;
  IplImage* cvColorImage = createDebugImage(parameters, color, edge);

  if (cvColorImage == NULL)
    {
      return VISION_MALLOC_ERROR;
    }

  int ndiv = 360; // 分割数
  double step = 2.0 * M_PI / ndiv;
  quaternion_t q;
  double genP[3];
  double dst[3];  // 円上の点（開始は３次元円周上の一点）
  Data_2D iPos;
  int sts, d;
  int ix, iy;
  CvPoint pt;
  CvScalar green = CV_RGB(0, 255, 0);
  StereoConic* conic;
  CircleCandidate* circle;
  int i;

  // 円の描画
  for (i = 0; i < stereo.numOfconics; i++)
    {
      conic = &stereo.conics[i];
      circle = &(stereo.conics[i].work.circle);
      if (conic->type != ConicType_Ellipse || circle->valid != 1)
        {
          continue;
        }

      // 円中心の描画
      projectXYZ2LR(&iPos, circle->center, (CameraParam*)cameraParam);
      ix = roundoff(iPos.col);
      iy = roundoff(iPos.row);
      pt = cvPoint(ix, iy);
      cvLine(cvColorImage, pt, pt, green);

      // 円の始点座標計算
      sts = getPointOnCircle(circle->normal, circle->radius, genP);
      // 始点計算ができないときは次の円へ
      if (sts)
        {
          continue;
        }

      // 回転を表す単位クォータニオンの計算
      quaternion_rotation(q, -step, circle->normal);
      for (d = 0; d < ndiv; d++)
        {
          // 円中心からの法線を回転軸にして円の点列座標を計算する
          quat_rot(genP, q, genP);
          addV3(genP, circle->center, dst);
          projectXYZ2LR(&iPos, dst, (CameraParam*)cameraParam);
          ix = roundoff(iPos.col);
          iy = roundoff(iPos.row);
          pt = cvPoint(ix, iy);
          // 円周の描画
          cvLine(cvColorImage, pt, pt, green);
        }
    }

  outDebugImage(cvColorImage, "circles3D", pairing, parameters.dbgdisp);

  return 0;
}

int
printStereoVertices(const StereoData& stereo, int pairing)
{
  // 頂点を構成する2ベクトルの保存
  // 出力結果 position, endpoint1, position, endpoint2 .....
  FILE* fp;
  int i;
  StereoConic* conic;
  VertexCandidate* vertex;
  char filename[PATH_MAX+1] = {0};

  snprintf(filename, sizeof(filename), "vertex3D%d.txt", pairing);
  fp = fopen(filename, "w");
  if (fp == NULL)
    {
      return VISION_FILE_OPEN_ERROR;
    }
  for (i = 0; i < stereo.numOfconics; i++)
    {
      conic = &stereo.conics[i];
      vertex = &conic->work.vertex;
      if (vertex->valid != 1)
        {
          continue;
        }
      fprintf(fp, "%f %f %f\n", vertex->position[0], vertex->position[1], vertex->position[2]);
      fprintf(fp, "%f %f %f\n", vertex->endpoint1[0], vertex->endpoint1[1], vertex->endpoint1[2]);
      fprintf(fp, "%f %f %f\n", vertex->position[0], vertex->position[1], vertex->position[2]);
      fprintf(fp, "%f %f %f\n", vertex->endpoint2[0], vertex->endpoint2[1], vertex->endpoint2[2]);
    }
  fclose(fp);
  return 0;
}

int
printStereoCircles(const StereoData& stereo, int pairing)
{
  // 円の保存
  FILE* fp;
  int i;
  StereoConic* conic;
  CircleCandidate* circle;
  double radius;
  double dst[3];
  int ndiv = 180; // 分割数
  double step = 2.0 * M_PI / ndiv; // 円の描画ステップ
  char filename[PATH_MAX+1] = {0};
  double genP[3];
  quaternion_t q;
  int sts, d;

  snprintf(filename, sizeof(filename), "circle3D%d.txt", pairing);
  fp = fopen(filename, "w");
  if (fp == NULL)
    {
      return VISION_FILE_OPEN_ERROR;
    }
  for (i = 0; i < stereo.numOfconics; i++)
    {
      conic = &stereo.conics[i];
      circle = &(stereo.conics[i].work.circle);
      if (conic->type != ConicType_Ellipse || circle->valid != 1)
        {
          continue;
        }

      // 円の始点座標計算
      sts = getPointOnCircle(circle->normal, circle->radius, genP);
      // 始点計算ができないときは次の円へ
      if (sts)
        {
          continue;
        }

      // 中心と法線
      radius = circle->radius;
      fprintf(fp, "%f %f %f\n", circle->center[0], circle->center[1], circle->center[2]);
      fprintf(fp, "%f %f %f\n", 
              circle->center[0] + radius * circle->normal[0], 
              circle->center[1] + radius * circle->normal[1],
              circle->center[2] + radius * circle->normal[2]);

      // 回転を表す単位クォータニオンの計算
      quaternion_rotation(q, -step, circle->normal);
      for (d = 0; d < ndiv; d++)
        {
          // 円中心からの法線を回転軸にして円の点列座標を計算する
          quat_rot(genP, q, genP);
          addV3(genP, circle->center, dst);
          fprintf(fp, "%f %f %f\n", dst[0], dst[1], dst[2]);
        }
    }
  fclose(fp);

  return 0;
}
